package net.tomatentum.cutin.method;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.tomatentum.cutin.container.MethodContainer;
import net.tomatentum.cutin.factory.ReflectedMethodFactory;
import net.tomatentum.cutin.factory.ReflectedMethodFactory.ParserResults;
import net.tomatentum.cutin.util.ReflectionUtil;

public abstract class BestCandidateMethod<I extends Object, C extends Object> extends ReflectedMethod<I, C> {

    private Logger logger = LoggerFactory.getLogger(getClass());

    private String methodName;
    
    protected BestCandidateMethod(String methodName, Object containingObject) {
        super(getMethod(containingObject, methodName), containingObject);
        this.methodName = methodName;
    }

    @Override
    public Object run(C context) {
        Method[] methods = Arrays.stream(containingObject.getClass().getDeclaredMethods())
            .filter(x -> x.getName().equals(methodName))
            .filter(x -> !x.isBridge())
            .toArray(Method[]::new);
        Class<?>[] parameters = getCurrentParameterList(context).stream()
            .map(Object::getClass)
            .toArray(Class<?>[]::new);
        super.method = ReflectionUtil.getMostSpecificMethod(methods, parameters);
        logger.trace("Found {} for {}({}) in {}", ReflectionUtil.getFullMethodName(method()), this.methodName, String.join(", ", Arrays.stream(parameters).map(Object::toString).toList()), this);
        return super.run(context);
    }

    private static Method getMethod(Object containingMethod, String methodName) {
        return Arrays.stream(containingMethod.getClass().getDeclaredMethods())
            .filter(m -> m.getName().equals(methodName))
            .findFirst().orElse(null);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof BestCandidateMethod<?, ?> bcMethod) {
            return this.containingObject().getClass().equals(bcMethod.containingObject().getClass()) &&
                this.methodName.equals(bcMethod.methodName);
        }
        return false;
    }

    @Override
    public int hashCode() {
        return Objects.hash(containingObject.getClass(), methodName);
    }

    private List<Object> getCurrentParameterList(C context) {
        List<Object> parameters = new ArrayList<>();
        int c = 0;
        Object last;
        while ((last = getParameter(context, c)) != null) {
            parameters.add(last);
            c++;
        }
        return parameters;
    }

    public abstract static class Factory<I extends Object, C extends Object> implements ReflectedMethodFactory.Factory<I, C> {

        private MethodContainer<I, C> methodContainer;
        private String methodName;

        protected Factory(MethodContainer<I, C> methodContainer, String methodName) {
            this.methodContainer = methodContainer;
            this.methodName = methodName;
        }

        @Override
        public Optional<ReflectedMethod<I, C>> produce(Method method, Object containingObject, ParserResults results) {
            Optional<BestCandidateMethod<I, C>> bcMethod = bcProduce(methodName, containingObject, results);
            if (bcMethod.isEmpty() || methodContainer.methods().contains(bcMethod.get()))
                return Optional.empty();
            return Optional.of(bcMethod.get());
        }

        protected abstract Optional<BestCandidateMethod<I, C>> bcProduce(String methodName, Object containingObject, ParserResults results);

    }
    
}
